﻿// BallPhysicsGame Version 1 - Using Papervision 2.0 Beta - Revision 870, JibLibFlash - Revision 25
// A very simple physics based ball game, where you must knock off red squares. Level information is loaded via an xml file
//
// released under MIT License
// http://www.opensource.org/licenses/mit-license.php
//
//@author		Devin Reimer
//@version		1.0.0
//@website 		http://www.permaven.com, http://www.almostlogical.com


//Copyright (c) 2009 Devin Reimer

/*
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

package com
{
	import flash.display.Shape;
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.events.KeyboardEvent;
	import flash.events.MouseEvent;
	import flash.text.TextField;
	import flash.text.TextFormat;
	import flash.ui.Keyboard;
	import flash.events.IOErrorEvent;
	import flash.display.Loader;
	import flash.net.URLLoader;
	import flash.net.URLRequest;
	import flash.text.TextField;
	import flash.text.TextFieldAutoSize;
	
	import org.papervision3d.core.math.Number3D;
	import org.ascollada.utils.FPS;
	import org.papervision3d.view.Viewport3D;
	import org.papervision3d.cameras.SpringCamera3D;
	import org.papervision3d.render.QuadrantRenderEngine;
	import org.papervision3d.materials.utils.MaterialsList;
	import org.papervision3d.materials.ColorMaterial;
	import org.papervision3d.materials.shadematerials.FlatShadeMaterial;
	import org.papervision3d.objects.primitives.Cube;
	import org.papervision3d.objects.primitives.Plane;
	import org.papervision3d.objects.primitives.Sphere;
	import org.papervision3d.objects.DisplayObject3D;
	import org.papervision3d.lights.PointLight3D;
	import org.papervision3d.scenes.Scene3D;
	import org.papervision3d.cameras.SpringCamera3D;

	import jiglib.math.*;
	import jiglib.geometry.*;
	import jiglib.physics.*;
	import jiglib.physics.constraint.*;


	public class BallPhysicsGame extends Sprite
	{
		private static const PLAYER_BALL_RADIUS:int = 25;
		private static const PLAYER_FORCE:int = 75;
		private static const PLAYER_WEIGHT:int = 5;
		private static const PLAYER_COLOUR:uint = 0xFBA248;
		
		private static const BASIC_GROUND_RESTITUTION:Number = 0.5;
		private static const BASIC_GROUND_STATIC_FRICTION:Number = 0.9;
		
		private static const WORLD_BOTTOM:int = -400;
		private static const STARTING_LEVEL:int = 1;
		private static const LEVEL_XML_PATH:String = "xml/LevelInfo.xml";
			
		private var viewport:Viewport3D;
		private var scene:Scene3D;
		private var camera:SpringCamera3D;
		private var renderer:QuadrantRenderEngine;
	
		private var lightSource:PointLight3D;
		
		private var currentLevel:int = STARTING_LEVEL;
		private var keyRight:Boolean = false;
		private var keyLeft:Boolean = false;
		private var keyForward:Boolean = false;
		private var keyBack:Boolean = false;
	
		private var xmlLoader:URLLoader
		
		private var playerBall:JSphere;
		private var playerCameraPosition:DisplayObject3D;
		private var cameraRotation:int = 0;
		private var isIntro:Boolean;
		private var introCounter:int;
	
		private var markerArray:Array;
		private var currentStatusTF:TextField;
		private var totalMarkers:int;
		
		private var instructionMessage:TextField;
		private var instructionFormat:TextFormat;

		public function BallPhysicsGame()
		{
			stage.addEventListener(MouseEvent.CLICK, setup, false, 0, true); //wait until the user click to start
			instructionMessage = new TextField();
			instructionFormat  = new TextFormat(null, 35, 0x000000);
			instructionMessage.defaultTextFormat = instructionFormat;
			instructionMessage.text = "Click to Start Game";
			instructionMessage.autoSize = TextFieldAutoSize.LEFT;
			instructionMessage.x= (stage.stageWidth - instructionMessage.width) / 2;
			instructionMessage.y = (stage.stageHeight / 2)-80;
			addChild(instructionMessage);
		}
		
		private function setup(evt:Event)
		{
			stage.removeEventListener(MouseEvent.CLICK, setup, false);
			removeChild(instructionMessage);
			stage.addEventListener( KeyboardEvent.KEY_DOWN, keyDownHandler );
			stage.addEventListener( KeyboardEvent.KEY_UP, keyUpHandler );
			isIntro = true; //is current in intro process
			introCounter = 0; //how many inc of rotation have occurred
			cameraRotation = 5; //the speed for camera rotation during intro
			
			initPapervision3D();
			createPlayer();
			playerCameraPosition = new DisplayObject3D();
			camera.target = playerCameraPosition;
			createHUD();
			loadLevelXML(LEVEL_XML_PATH);
			
			//addChild(new FPS());
		}

		private function initPapervision3D():void
		{
			viewport = new Viewport3D(550,400);
			addChild(viewport);
			
			scene = new Scene3D();
			 
			lightSource = new PointLight3D(true, true);
			lightSource.x = 100;
			lightSource.y = 300;
			lightSource.z = -500;
					 			 
			camera = new SpringCamera3D();
			camera.mass = 10;
			camera.damping = 6;
			camera.stiffness = 3;
				
			camera.lookOffset = new Number3D(0, 20, 30);
			camera.positionOffset = new Number3D(0, 100, -270);
			
			camera.focus = 100;
			camera.zoom = 10;
			 
			renderer = new QuadrantRenderEngine(QuadrantRenderEngine.CORRECT_Z_FILTER);
		}
		
		private function createPlayer():void
		{
			var ballMaterial = new FlatShadeMaterial(lightSource,PLAYER_COLOUR);
			var ballSkin:Sphere = new Sphere(ballMaterial,PLAYER_BALL_RADIUS);
			
			scene.addChild(ballSkin);
			playerBall = new JSphere(ballSkin, true, PLAYER_BALL_RADIUS);
			playerBall.setMass(PLAYER_WEIGHT);
			resetPlayerPosition();
			PhysicsSystem.getInstance().AddBody(playerBall);
		}
		
		//loads in an xml to build the level
		private function buildLevelUsingXML(theXML:XML):void
		{
			var levelXML:XMLList;
			
			levelXML = theXML.level.(@id == currentLevel);
			
			markerArray = new Array();
			
			for each (var staticObjXML:XML in levelXML.staticObj)
			{
				createStaticObjFromXML(staticObjXML);
			}
			
			for each (var physicsObjXML:XML in levelXML.physicsObj)
			{
				createPhysicsObjFromXML(physicsObjXML);
			}
			
			totalMarkers = markerArray.length;
			beginRender();
		}
		
		private function createStaticObjFromXML(staticObjXML:XML):void
		{
			var basicPropXML:XMLList = staticObjXML.basicProp;
			var groundMatList:MaterialsList = new MaterialsList();
			var groundSkin:Cube;
			var groundColour:uint = staticObjXML.@colour;
			var platform:JBox;
			var platformWidth:int = basicPropXML.@width;
			var platformDepth:int = basicPropXML.@depth;
			var platformHeight:int = basicPropXML.@height;
			var platformX:int = basicPropXML.@x;
			var platformY:int = basicPropXML.@y;
			var platformZ:int = basicPropXML.@z;
		
			groundMatList.addMaterial(new FlatShadeMaterial(lightSource,groundColour),"all");
			groundSkin = new Cube(groundMatList, platformWidth, platformDepth, platformHeight,3,3,3);
			
			scene.addChild(groundSkin);
			platform = new JBox(groundSkin, false, platformWidth, platformDepth, platformHeight);
			platform.MoveTo(new JNumber3D(platformX,platformY,platformZ), JMatrix3D.IDENTITY);
	
			
			PhysicsSystem.getInstance().AddBody(platform);
			platform.Material.Restitution = BASIC_GROUND_RESTITUTION;
			platform.Material.StaticFriction = BASIC_GROUND_STATIC_FRICTION;
		
		}
		
		private function createPhysicsObjFromXML(physicsObjXML:XML):void
		{
			var basicPropXML:XMLList = physicsObjXML.basicProp;
			var objMatList:MaterialsList = new MaterialsList();
			var objSkin:Cube;
			var objColour:uint = physicsObjXML.@colour;
			var physicsObj:JBox;
			var physicsObjWidth:int = basicPropXML.@width;
			var physicsObjDepth:int = basicPropXML.@depth;
			var physicsObjHeight:int = basicPropXML.@height;
			var physicsObjX:int = basicPropXML.@x;
			var physicsObjY:int = basicPropXML.@y;
			var physicsObjZ:int = basicPropXML.@z;
			var physicsObjMass:int = physicsObjXML.@mass;
			
			objMatList.addMaterial(new FlatShadeMaterial(lightSource,objColour),"all");
			objSkin = new Cube(objMatList, physicsObjWidth, physicsObjDepth, physicsObjHeight);
			
			scene.addChild(objSkin);
			
			physicsObj = new JBox(objSkin, true, physicsObjWidth, physicsObjDepth, physicsObjHeight);
			physicsObj.MoveTo(new JNumber3D(physicsObjX, physicsObjY, physicsObjZ), JMatrix3D.IDENTITY);
			physicsObj.setMass(physicsObjMass);
			
			if (physicsObjXML.@type == "markerbox")
			{
				markerArray.push(physicsObj);
			}
			
			PhysicsSystem.getInstance().AddBody(physicsObj);
		}
		
		private function loadLevelXML(xmlFileName:String):void
		{
			xmlLoader = new URLLoader();
			xmlLoader.load(new URLRequest(xmlFileName));
			xmlLoader.addEventListener(Event.COMPLETE,onXMLLoadComplete,false,0,true);
			xmlLoader.addEventListener(IOErrorEvent.IO_ERROR, onLoadFailure,false,0,true);
		}
			
		private function onXMLLoadComplete(evt:Event):void
		{
			var theXML:XML;
			try
			{
				XML.ignoreWhitespace = true;
				theXML = XML(evt.target.data);
			}
			catch(error:Error)
			{
				trace("Error In XML (General)- " + error.message);
			}
			
			xmlLoader.removeEventListener(Event.COMPLETE,onXMLLoadComplete,false);
			xmlLoader.removeEventListener(IOErrorEvent.IO_ERROR, onLoadFailure, false);
			xmlLoader = null;
			
			buildLevelUsingXML(theXML);
			
		}
		private function onLoadFailure(evt:Event):void
		{
			trace("Load Failure: " + evt);
		}

		private function keyDownHandler(evt:KeyboardEvent):void
		{
			switch(evt.keyCode)
			{
				case Keyboard.UP:
					keyForward = true;
					keyBack = false;
					break;
				case Keyboard.DOWN:
					keyBack = true;
					keyForward = false;
					break;
				case Keyboard.LEFT:
					keyLeft = true;
					keyRight = false;
					break;
				case Keyboard.RIGHT:
					keyRight = true;
					keyLeft = false;
					break;
				case Keyboard.SPACE:
					trace(playerBall.CurrentState.Position.toString());
					break;
			}
		}

		private function keyUpHandler(evt:KeyboardEvent):void
		{
			switch(evt.keyCode)
			{
				case Keyboard.UP:
					keyForward = false;
					break;
				case Keyboard.DOWN:
					keyBack = false;
					break;
				case Keyboard.LEFT:
					keyLeft = false;
					break;
				case Keyboard.RIGHT:
					keyRight = false;
					break;
			}
		}
		
		private function checkPlayerPosition():void
		{
			if (playerBall.CurrentState.Position.y < WORLD_BOTTOM) //if fall off level reset layer
			{
				resetPlayerPosition();
			}
		}
		
		private function resetPlayerPosition():void
		{
			playerBall.MoveTo(new JNumber3D(0,PLAYER_BALL_RADIUS,0), JMatrix3D.IDENTITY);
		}

		private function beginRender():void
		{
			updateHUDText();
			this.addEventListener(Event.ENTER_FRAME,render);
		}
		private function render(evt:Event):void
		{			
			movePlayer();
			PhysicsSystem.getInstance().Integrate(0.1);
			checkPlayerPosition();
			checkMarkers();
			updateCamera();
	
			renderer.renderScene( scene, camera, viewport );
		}
		
		private function updateCamera():void
		{
			//this following code is for initial spin around the target ball
			if (isIntro && introCounter>=360/cameraRotation)
			{
				isIntro = false;
				cameraRotation = 0;
				playerCameraPosition.rotationY = 0;
				removeChild(instructionMessage);
			}
			else if (isIntro)
			{
				introCounter++;
				playerCameraPosition.yaw(cameraRotation);
			}
			
			playerCameraPosition.copyPosition(JSphere(playerBall).BodySkin); //update camera position excluding angle
		}
		
		private function movePlayer():void
		{
			var xForce:int = 0;
			var yForce:int = 0;
			var zForce:int = 0;
	
			if(keyLeft)
			{
				xForce-= PLAYER_FORCE;
			}
			if(keyRight)
			{
				xForce += PLAYER_FORCE;
			}
			if(keyForward)
			{
				zForce += PLAYER_FORCE;
			}
			if(keyBack)
			{
				zForce -= PLAYER_FORCE;
			}
			
			playerBall.AddWorldForce(new JNumber3D(xForce,yForce,zForce),playerBall.CurrentState.Position);
		}
		
		
		private function checkMarkers()
		{
			var markerObj:JBox;
			var isUpdate:Boolean = false;
			
			for (var counter:int = 0; counter < markerArray.length; counter++)
			{
				markerObj = markerArray[counter];
				if (markerObj != null)
				{
					if (markerObj.CurrentState.Position.y < WORLD_BOTTOM) 
					{
						PhysicsSystem.getInstance().RemoveBody(markerObj);
						scene.removeChild(markerObj.BodySkin);
						markerArray.splice(markerArray.indexOf(markerObj), 1);
						counter--;
						markerObj = null;
						isUpdate = true;
					}
				}
			}
			
			if (isUpdate)
			{
				updateHUDText();
			}
		}
		
		private function createHUD()
		{
			var backgroundSprite:Sprite = new Sprite();
			var blogNameTF:TextField;
			
			backgroundSprite.graphics.beginFill(0xFFFFFF);
			backgroundSprite.graphics.drawRect(0, 0, 200, 20);
			backgroundSprite.graphics.endFill();
			
			currentStatusTF = new TextField();
		
			currentStatusTF.text = "Loading...";
			currentStatusTF.textColor = 0xF70000;
			currentStatusTF.autoSize = TextFieldAutoSize.LEFT;
			currentStatusTF.x = 20;
			currentStatusTF.y = 2;
			backgroundSprite.addChild(currentStatusTF);
			
			backgroundSprite.x = (stage.stageWidth - backgroundSprite.width);
			addChild(backgroundSprite);
			
			instructionMessage = new TextField();
			instructionFormat  = new TextFormat(null, 35, 0xF70000);
			instructionMessage.defaultTextFormat = instructionFormat;
			instructionMessage.text = "Knock off all red boxes to win!";
			instructionMessage.autoSize = TextFieldAutoSize.LEFT;
			instructionMessage.x= (stage.stageWidth - instructionMessage.width) / 2;
			//instructionMessage.width = stage.stageWidth;
			instructionMessage.y = (stage.stageHeight / 2)-80;
			addChild(instructionMessage);
			
			blogNameTF = new TextField();
			blogNameTF.text = "http://blog.almostlogical.com - Devin Reimer";
			blogNameTF.textColor = 0xFFFFFF;
			blogNameTF.autoSize = TextFieldAutoSize.LEFT;
			blogNameTF.x = (stage.stageWidth - blogNameTF.width);
			blogNameTF.y = (stage.stageHeight - blogNameTF.height);
			addChild(blogNameTF);
		}
		
		private function updateHUDText()
		{
			if (markerArray.length == 0)
			{
				currentStatusTF.text = "Winner!!!";
				instructionMessage.text = "Winner!!!";
				instructionMessage.autoSize = TextFieldAutoSize.LEFT;
				instructionMessage.x= (stage.stageWidth - instructionMessage.width) / 2;
				addChild(instructionMessage);
			}
			else
			{
				currentStatusTF.text = "Red Boxes Remaining: " + markerArray.length + " out of " + totalMarkers;
			}
		}
	}//end class BallPhysics
}